@@ -1,7 +1,7 @@
 /* vim: set expandtab ts=4 sw=4: */
 /*
  *
- *  Copyright © 2006 Juha Kautto <juha@xfce.org>
+ *  Copyright © 2006-2007 Juha Kautto <juha@xfce.org>
  *
  *  it under the terms of the GNU Library General Public License as published 
  *  by the Free Software Foundation; either version 2 of the License, or
@@ -44,86 +44,260 @@
  *                               Clock                                  *
  * -------------------------------------------------------------------- */
 
-
-static gboolean oc_date_tooltip(Clock *clock)
+static void utf8_strftime(char *res, int res_l, char *format, struct tm *tm)
 {
-    char date_s[255];
-    char *utf8date = NULL;
+    char *tmp = NULL;
 
-    /* TRANSLATORS: Use format characters from strftime(3)
-     * to get the proper string for your locale.
-     * I used these:
-     * %A  : full weekday name
-     * %d  : day of the month
-     * %B  : full month name
-     * %Y  : four digit year
-     * %V  : ISO week number
-     */
-    strftime(date_s, 255
-            , g_locale_from_utf8( _("%A %d %B %Y/%V"), -1, NULL, NULL, NULL)
-            , &clock->now);
-
-    /* Conversion to utf8 */
-    if (!g_utf8_validate(date_s, -1, NULL)) {
-        utf8date = g_locale_to_utf8(date_s, -1, NULL, NULL, NULL);
+    /* strftime is nasty. It returns formatted characters (%A...) in utf8
+     * but it does not convert plain characters so they will be in locale 
+     * charset. 
+     * It expects format to be in locale charset, so we need to convert 
+     * that first (it may contain utf8).
+     * We need then convert the results finally to utf8.
+     * */
+    tmp = g_locale_from_utf8(format, -1, NULL, NULL, NULL);
+    strftime(res, res_l, tmp, tm);
+    g_free(tmp);
+    /* Then convert to utf8 if needed */
+    if (!g_utf8_validate(res, -1, NULL)) {
+        tmp = g_locale_to_utf8(res, -1, NULL, NULL, NULL);
+        if (tmp) {
+            g_strlcpy(res, tmp, res_l);
+            g_free(tmp);
+        }
     }
+}
 
-    if (utf8date) {
-        gtk_tooltips_set_tip(clock->tips, GTK_WIDGET(clock->plugin), 
-                utf8date, NULL);
-        g_free(utf8date);
-    }
-    else {
-        gtk_tooltips_set_tip(clock->tips, GTK_WIDGET(clock->plugin), 
-                date_s, NULL);
-    }
+static void oc_tooltip_set(Clock *clock)
+{
+    char res[OC_MAX_LINE_LENGTH-1];
 
-    return(TRUE);
+    utf8_strftime(res, sizeof(res), clock->tooltip_data->str, &clock->now);
+    if (strcmp(res,  clock->tooltip_prev)) {
+        gtk_tooltips_set_tip(clock->tips, GTK_WIDGET(clock->plugin),res, NULL);
+        strcpy(clock->tooltip_prev, res);
+    }
 }
 
 static gboolean oc_get_time(Clock *clock)
 {
     time_t  t;
-    char    time_s[OC_MAX_LINE_LENGTH-1];
-    char    *utf8time_s = NULL;
+    char    res[OC_MAX_LINE_LENGTH-1];
     int     i;
-    static gint mday = -1;
     ClockLine *line;
 
     time(&t);
     localtime_r(&t, &clock->now);
-
     for (i = 0; i < OC_MAX_LINES; i++) {
         line = &clock->line[i];
         if (line->show) {
-            strftime(time_s, sizeof(time_s)
-                    , g_locale_from_utf8(line->data->str, -1, NULL, NULL, NULL)
-                    , &clock->now);
-            if (!g_utf8_validate(time_s, -1, NULL)) {
-                utf8time_s = g_locale_to_utf8(time_s, -1, NULL, NULL, NULL);
-                if (utf8time_s) {
-                    g_strlcpy(time_s, utf8time_s, sizeof(time_s));
-                    g_free(utf8time_s);
-                }
-            }
+            utf8_strftime(res, sizeof(res), line->data->str, &clock->now);
             /* gtk_label_set_text call takes almost
              * 100 % of the time wasted in this procedure 
+             * Note that even though we only wake up when needed, we 
+             * may not have to update all lines, so this check still
+             * probably is worth doing
              * */
-            if (strcmp(time_s,  line->prev)) {
-                gtk_label_set_text(GTK_LABEL(line->label), time_s);
-                strcpy(line->prev, time_s);
+            if (strcmp(res,  line->prev)) {
+                gtk_label_set_text(GTK_LABEL(line->label), res);
+                strcpy(line->prev, res);
             }
         }
     }
-
-    if (mday != clock->now.tm_mday) {
-        oc_date_tooltip(clock);
-        mday = clock->now.tm_mday;
-    }
+    oc_tooltip_set(clock);
 
     return(TRUE);
 }
 
+static gboolean oc_get_time_delay(Clock *clock)
+{
+    clock->timeout_id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE
+            , clock->interval, (GSourceFunc)oc_get_time, clock, NULL);
+    oc_get_time(clock);
+    return(FALSE); /* this is one time only timer */
+}
+
+gboolean oc_start_timer(Clock *clock)
+{
+    time_t t;
+    gint   delay_time =  0;
+    /* if we are using longer than 1 minute (= 60000) interval, we need
+     * to delay the first start so that clock changes when minute or hour
+     * changes */
+
+    oc_get_time(clock);
+    time(&t);
+    localtime_r(&t, &clock->now);
+    if (clock->interval >= 60000) {
+        if (clock->interval >= 3600000) /* match to next full hour */
+            delay_time = (clock->interval -
+                    (clock->now.tm_min*60000 + clock->now.tm_sec*1000));
+        else /* match to next full minute */
+            delay_time = (clock->interval - clock->now.tm_sec*1000);
+    }
+    if (clock->delay_timeout_id) {
+        g_source_remove(clock->delay_timeout_id);
+        clock->delay_timeout_id = 0;
+    }
+    if (clock->timeout_id) {
+        g_source_remove(clock->timeout_id);
+        clock->timeout_id = 0;
+    }
+    clock->delay_timeout_id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE
+            , delay_time, (GSourceFunc)oc_get_time_delay, clock, NULL);
+    /* if we have longer than 1 sec timer, we need to reschedule 
+     * it regularly since it will fall down slowly but surely, so
+     * we keep this running. */
+    if (clock->interval >= 60000) {
+        if (delay_time > 60000)
+        /* let's run it once in case we happened to kill it
+           just when it was supposed to start */
+            oc_get_time(clock); 
+        return(FALSE);
+    }
+    else
+        return(TRUE);
+}
+
+static gboolean oc_end_tuning(Clock *clock)
+{
+    /* if we have longer than 1 sec timer, we need to reschedule 
+     * it regularly since it will fall down slowly but surely */
+    if (clock->adjust_timeout_id) {
+        g_source_remove(clock->adjust_timeout_id);
+        clock->adjust_timeout_id = 0;
+    }
+    if (clock->interval >= 60000) { /* resync it after each 6 hours */
+        clock->adjust_timeout_id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE
+                , 6*60*60*1000, (GSourceFunc)oc_start_timer, clock, NULL);
+    }
+    g_free(clock->tune);
+    clock->tune = NULL;
+}
+
+static gboolean oc_next_level_tuning(Clock *clock)
+{
+    TimeTuning *tune = clock->tune;
+    gboolean continue_tuning = TRUE;
+
+    if (clock->interval == OC_CONFIG_INTERVAL) /* setup active => no changes */
+        return(FALSE);
+    switch (tune->level) {
+        case 0: /* second now; let's set interval to minute */
+            clock->interval = 60*1000; 
+            break;
+        case 1: /* minute now; let's set it to hour */
+            clock->interval = 60*60*1000; 
+            break;
+        default: /* no point tune more */
+            break;
+    }
+    if (tune->level < 2) {
+        tune->level++;
+        oc_start_timer(clock);
+        oc_start_tuning(clock);
+    }
+    else {
+        continue_tuning = FALSE;
+        oc_end_tuning(clock);
+    }
+    return(continue_tuning);
+}
+
+static gboolean oc_tune_time(Clock *clock)
+{
+    /* idea is that we compare the panel string three times and 
+     * if all of those are the same, we can tune the timer since we know 
+     * that we do not have to check it so often. Time in panel plugin is
+     * not changing anyway. So we adjust the timers and check again */
+    gboolean continue_tuning = TRUE;
+    TimeTuning *tune = clock->tune;
+
+    if (clock->interval == OC_CONFIG_INTERVAL) /* setup active => no changes */
+        return(FALSE);
+    switch (tune->cnt) {
+        case 0: /* first time we come here */
+            strncpy(tune->prev[0], clock->line[0].prev, OC_MAX_LINE_LENGTH);
+            strncpy(tune->prev[1], clock->line[1].prev, OC_MAX_LINE_LENGTH);
+            strncpy(tune->prev[2], clock->line[2].prev, OC_MAX_LINE_LENGTH);
+            strncpy(tune->tooltip_prev, clock->tooltip_prev,OC_MAX_LINE_LENGTH);
+            break;
+        case 1:
+        case 2:
+        case 3: /* FIXME: two may be enough ?? */
+            if (strncmp(tune->prev[0], clock->line[0].prev, OC_MAX_LINE_LENGTH)
+            ||  strncmp(tune->prev[1], clock->line[1].prev, OC_MAX_LINE_LENGTH)
+            ||  strncmp(tune->prev[2], clock->line[2].prev, OC_MAX_LINE_LENGTH)
+            ||  strncmp(tune->tooltip_prev, clock->tooltip_prev
+                    ,OC_MAX_LINE_LENGTH)) {
+                /* time has changed. But it may change due to bigger time
+                 * change. For example if we are checking seconds and time
+                 * only shows minutes, this can change if hour happens to
+                 * change. So we need to check the change twice to be sure */
+                if (tune->changed_once) { /* this is the second trial already */
+                    continue_tuning = FALSE;
+                    oc_end_tuning(clock);
+                }
+                else { /* need to try again */
+                    tune->changed_once = TRUE;
+                    tune->cnt = -1; /* this gets incremented to 0 below */
+                }
+            }
+            break;
+        default:
+            break;
+    }
+    if (continue_tuning)
+        if (tune->cnt == 3) 
+            continue_tuning = oc_next_level_tuning(clock);
+        else 
+            tune->cnt++;
+
+    return(continue_tuning);
+}
+
+void oc_disable_tuning(Clock *clock)
+{
+    TimeTuning *tune = clock->tune;
+
+    if (tune == NULL)
+        return;
+    if (tune->timeout_id) {
+        g_source_remove(tune->timeout_id);
+        tune->timeout_id = 0;
+    }
+    if (clock->adjust_timeout_id) {
+        g_source_remove(clock->adjust_timeout_id);
+        clock->adjust_timeout_id = 0;
+    }
+}
+
+void oc_start_tuning(Clock *clock)
+{
+    TimeTuning *tune = clock->tune;
+
+    if (clock->interval == OC_CONFIG_INTERVAL) /* setup active => no changes */
+        return;
+    tune->cnt = 0;
+    tune->changed_once = FALSE;
+    if (tune->timeout_id) {
+        g_source_remove(tune->timeout_id);
+        tune->timeout_id = 0;
+    }
+    tune->timeout_id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE
+            , 2*clock->interval, (GSourceFunc)oc_tune_time, clock, NULL);
+}
+
+void oc_init_timer(Clock *clock)
+{
+    clock->interval = OC_BASE_INTERVAL;
+    if (clock->tune == NULL)
+        clock->tune = g_new0(TimeTuning, 1);
+    clock->tune->level = 0;
+    oc_start_timer(clock);
+    oc_start_tuning(clock);
+}
+
 static void oc_update_size(Clock *clock, int size)
 {
     /* keep in sync with systray */
@@ -137,7 +311,8 @@ static void oc_update_size(Clock *clock, int size)
     }
 }
 
-static gboolean popup_program(GtkWidget *widget, gchar *program, Clock *clock)
+static gboolean popup_program(GtkWidget *widget, gchar *program, Clock *clock
+        , guint event_time)
 {
     GdkAtom atom;
     Window xwindow;
@@ -154,7 +329,7 @@ static gboolean popup_program(GtkWidget *widget, gchar *program, Clock *clock)
         popup = "_XFCE_GLOBALTIME_TOGGLE_HERE";
     }
     else {
-        g_warning("%s: unknown program to start %s", OC_NAME, program);
+        g_warning("unknown program to start %s", program);
         return(FALSE);
     }
 
@@ -176,6 +351,14 @@ static gboolean popup_program(GtkWidget *widget, gchar *program, Clock *clock)
         return(TRUE);
     }
     else { /* not running, let's try to start it. Need to reset TZ! */
+        static guint prev_event_time = 0; /* prevenst double start (BUG 4096) */
+
+        if (prev_event_time && ((event_time - prev_event_time) < 1000)) {
+            g_message("%s: double start of %s prevented", OC_NAME, program);
+            return(FALSE);
+        }
+            
+        prev_event_time = event_time;
         if (clock->TZ_orig != NULL)  /* we had TZ when we started */
             g_setenv("TZ", clock->TZ_orig, 1);
         else  /* TZ was not set so take it out */
@@ -200,10 +383,12 @@ static gboolean popup_program(GtkWidget *widget, gchar *program, Clock *clock)
 static gboolean on_button_press_event_cb(GtkWidget *widget
         , GdkEventButton *event, Clock *clock)
 {
+    if (event->type != GDK_BUTTON_PRESS) /* double or triple click */
+        return(FALSE); /* ignore */
     if (event->button == 1)
-        return(popup_program(widget, "orage", clock));
+        return(popup_program(widget, "orage", clock, event->time));
     else if (event->button == 2)
-        return(popup_program(widget, "globaltime", clock));
+        return(popup_program(widget, "globaltime", clock, event->time));
 
     return(FALSE);
 }
@@ -230,7 +415,9 @@ static void oc_free_data(XfcePanelPlugin *plugin, Clock *clock)
     if (dlg)
         gtk_widget_destroy(dlg);
     
-    g_source_remove(clock->timeout_id);
+    if (clock->timeout_id) {
+        g_source_remove(clock->timeout_id);
+    }
     g_object_unref(clock->tips);
     g_object_unref(clock->line[0].label);
     g_object_unref(clock->line[1].label);
@@ -247,12 +434,10 @@ static void oc_read_rc_file(XfcePanelPlugin *plugin, Clock *clock)
     gchar tmp[100];
     gint i;
 
-    if (!(file = xfce_panel_plugin_lookup_rc_file(plugin))) {
-        g_warning("%s: unable to read rc file", OC_NAME);
-        return;
-    }
+    if (!(file = xfce_panel_plugin_lookup_rc_file(plugin)))
+        return; /* if it does not exist, we use defaults from orage_oc_new */
     if (!(rc = xfce_rc_simple_open(file, TRUE))) {
-        g_warning("%s: unable to read-open rc file (%s)", OC_NAME, file);
+        g_warning("unable to read-open rc file (%s)", file);
         return;
     }
     g_free(file);
@@ -305,6 +490,9 @@ static void oc_read_rc_file(XfcePanelPlugin *plugin, Clock *clock)
         }
     }
 
+    if (ret = xfce_rc_read_entry(rc, "tooltip", NULL))
+        g_string_assign(clock->tooltip_data, ret); 
+
     xfce_rc_close(rc);
 }
 
@@ -316,11 +504,11 @@ void oc_write_rc_file(XfcePanelPlugin *plugin, Clock *clock)
     gint i;
 
     if (!(file = xfce_panel_plugin_save_location(plugin, TRUE))) {
-        g_warning("%s: unable to write rc file", OC_NAME);
+        g_warning("unable to write rc file");
         return;
     }
     if (!(rc = xfce_rc_simple_open(file, FALSE))) {
-        g_warning("%s: unable to read-open rc file (%s)", OC_NAME, file);
+        g_warning("unable to read-open rc file (%s)", file);
         return;
     }
     g_free(file);
@@ -382,6 +570,8 @@ void oc_write_rc_file(XfcePanelPlugin *plugin, Clock *clock)
         }
     }
 
+    xfce_rc_write_entry(rc, "tooltip",  clock->tooltip_data->str);
+
     xfce_rc_close(rc);
 }
 
@@ -410,6 +600,8 @@ Clock *orage_oc_new(XfcePanelPlugin *plugin)
     clock->show_frame = TRUE;
     clock->fg_set = FALSE;
     clock->bg_set = FALSE;
+    clock->width_set = FALSE;
+    clock->height_set = FALSE;
 
     clock->timezone = g_string_new(""); /* = not set */
     clock->TZ_orig = g_strdup(g_getenv("TZ"));
@@ -423,12 +615,21 @@ Clock *orage_oc_new(XfcePanelPlugin *plugin)
         clock->line[i].font = g_string_new("");
     }
 
+    /* TRANSLATORS: Use format characters from strftime(3)
+     * to get the proper string for your locale.
+     * I used these:
+     * %A  : full weekday name
+     * %d  : day of the month
+     * %B  : full month name
+     * %Y  : four digit year
+     * %V  : ISO week number
+     */
+    clock->tooltip_data = g_string_new(_("%A %d %B %Y/%V"));
+
     clock->tips = gtk_tooltips_new();
     g_object_ref(clock->tips);
     gtk_object_sink(GTK_OBJECT(clock->tips));
-    oc_get_time(clock);
-    clock->timeout_id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE
-            , 200, (GSourceFunc) oc_get_time, clock, NULL);
+    oc_init_timer(clock);
         
     return(clock);
 }
